use tokio::signal::ctrl_c;

mod health;
mod messaging;
mod pty;

#[tokio::main]
async fn main() -> anyhow::Result<()> {
    let subscriber = tracing_subscriber::fmt()
        .with_max_level(tracing::Level::TRACE)
        .finish();

    tracing::subscriber::set_global_default(subscriber)
        .expect("to set a tracing global subscriber");

    run().await.map_err(|err| {
        tracing::error!(%err, "failed to start the agent!");
        err.context("failed to start the Prymn agent")
    })
}

async fn run() -> anyhow::Result<()> {
    // TODO: Configurable client
    let client = messaging::Client::connect("1").await?;

    init_health_subsystem(client.clone());
    tracing::info!("initialized health subsystem");

    crate::messaging::init_services(client).await?;
    tracing::info!("initialized services");

    ctrl_c().await?;
    Ok(())
}

fn init_health_subsystem(client: messaging::Client) {
    let health = health::HealthMonitor::new();
    let mut health_monitor = health.monitor();

    tokio::spawn(async move {
        while health_monitor.changed().await.is_ok() {
            let data = {
                let health = &*health_monitor.borrow();
                tracing::info!(?health, "health watermark");
                serde_json::to_vec(health).unwrap()
            };

            if let Err(err) = client.publish("health", data).await {
                // TODO: Error recovery?
                tracing::warn!(cause = %err, "health subsystem publishing failed");
            }
        }
    });

    std::thread::spawn(move || {
        const INTERVAL: std::time::Duration = std::time::Duration::from_secs(5);
        let mut system = health::System::new();

        loop {
            health.check_system(&mut system);
            std::thread::sleep(INTERVAL);
        }
    });
}
