use std::{io, task::ready};

use rustix::{
    fd::OwnedFd,
    fs::{fcntl_getfl, fcntl_setfl, OFlags},
    process::{ioctl_tiocsctty, setsid},
    pty::{grantpt, ioctl_tiocgptpeer, openpt, unlockpt, OpenptFlags},
    stdio::{dup2_stderr, dup2_stdin, dup2_stdout},
    termios::{tcsetwinsize, Winsize},
};
use tokio::{
    io::{unix::AsyncFd, AsyncRead, AsyncWrite},
    process::Child,
};

#[derive(Debug)]
pub struct Pty {
    fd: AsyncFd<OwnedFd>,
}

impl Pty {
    pub fn open() -> io::Result<Self> {
        let master = openpt(OpenptFlags::RDWR | OpenptFlags::NOCTTY | OpenptFlags::CLOEXEC)?;
        grantpt(&master)?;
        unlockpt(&master)?;

        // Set nonblocking
        let flags = fcntl_getfl(&master)?;
        fcntl_setfl(&master, flags | OFlags::NONBLOCK)?;

        let fd = AsyncFd::new(master)?;

        Ok(Self { fd })
    }

    pub fn child(&self) -> io::Result<PtyChild> {
        // NOTE: Linux v4.13 and above
        let fd = ioctl_tiocgptpeer(&self.fd, OpenptFlags::RDWR | OpenptFlags::NOCTTY)?;
        let child = PtyChild { fd };

        Ok(child)
    }

    pub fn resize_window(&self, rows: u16, cols: u16) -> io::Result<()> {
        let winsize = Winsize {
            ws_row: rows,
            ws_col: cols,
            ws_xpixel: 0,
            ws_ypixel: 0,
        };

        tcsetwinsize(&self.fd, winsize)?;

        Ok(())
    }

    pub fn try_clone(&self) -> io::Result<Pty> {
        let fd = self.fd.get_ref().try_clone()?;

        Ok(Pty {
            fd: AsyncFd::new(fd)?,
        })
    }
}

impl AsyncRead for Pty {
    fn poll_read(
        self: std::pin::Pin<&mut Self>,
        cx: &mut std::task::Context<'_>,
        buf: &mut tokio::io::ReadBuf<'_>,
    ) -> std::task::Poll<io::Result<()>> {
        loop {
            let mut guard = ready!(self.fd.poll_read_ready(cx)?);

            match guard.try_io(|inner| {
                let fd = inner.get_ref();
                let n = rustix::io::read(fd, buf.initialize_unfilled())?;
                buf.advance(n);

                Ok(())
            }) {
                Ok(result) => return std::task::Poll::Ready(result),
                Err(_would_block) => continue,
            }
        }
    }
}

impl AsyncWrite for Pty {
    fn poll_write(
        self: std::pin::Pin<&mut Self>,
        cx: &mut std::task::Context<'_>,
        buf: &[u8],
    ) -> std::task::Poll<Result<usize, io::Error>> {
        loop {
            let mut guard = ready!(self.fd.poll_write_ready(cx))?;

            match guard.try_io(|inner| Ok(rustix::io::write(inner.get_ref(), buf)?)) {
                Ok(result) => return std::task::Poll::Ready(result),
                Err(_would_block) => continue,
            }
        }
    }

    fn poll_flush(
        self: std::pin::Pin<&mut Self>,
        _cx: &mut std::task::Context<'_>,
    ) -> std::task::Poll<Result<(), io::Error>> {
        std::task::Poll::Ready(Ok(()))
    }

    fn poll_shutdown(
        self: std::pin::Pin<&mut Self>,
        _cx: &mut std::task::Context<'_>,
    ) -> std::task::Poll<Result<(), io::Error>> {
        std::task::Poll::Ready(Ok(()))
    }
}

#[derive(Debug)]
pub struct PtyChild {
    fd: OwnedFd,
}

impl PtyChild {
    pub fn login_tty(&self) -> io::Result<()> {
        setsid()?;
        ioctl_tiocsctty(&self.fd)?;
        dup2_stdin(&self.fd)?;
        dup2_stdout(&self.fd)?;
        dup2_stderr(&self.fd)?;

        Ok(())
    }
}

pub fn open_shell(pty_child: PtyChild, shell: &str) -> io::Result<Child> {
    let mut cmd = tokio::process::Command::new(shell);

    unsafe {
        cmd.pre_exec(move || {
            pty_child.login_tty()?;
            Ok(())
        });
    }

    cmd.spawn()
}

#[cfg(test)]
mod test {
    use rustix::fd::AsRawFd;

    use super::*;

    #[tokio::test]
    async fn can_open_pty() {
        let pty = Pty::open().unwrap();
        let child = pty.child().unwrap();

        let master_fd = pty.fd.get_ref().as_raw_fd();
        let child_fd = child.fd.as_raw_fd();

        assert!(master_fd != child_fd);
    }
}
